"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var process = require("process");
var getStream = require("get-stream");
var _ = require("lodash");
var immutable_1 = require("immutable");
var targetLanguages = require("./Language/All");
var Annotation_1 = require("./Annotation");
var Support_1 = require("./Support");
var CompressedJSON_1 = require("./CompressedJSON");
var URLGrammar_1 = require("./URLGrammar");
var CombineClasses_1 = require("./CombineClasses");
var JSONSchemaInput_1 = require("./JSONSchemaInput");
var Inference_1 = require("./Inference");
var InferMaps_1 = require("./InferMaps");
var TypeBuilder_1 = require("./TypeBuilder");
var commandLineArgs = require("command-line-args");
var getUsage = require("command-line-usage");
var fetch = require("node-fetch");
var chalk = require("chalk");
var stringToStream = require("string-to-stream");
var langs = targetLanguages.all.map(function (r) { return _.minBy(r.names, function (s) { return s.length; }); }).join("|");
var langDisplayNames = targetLanguages.all.map(function (r) { return r.displayName; }).join(", ");
var optionDefinitions = [
    {
        name: "out",
        alias: "o",
        type: String,
        typeLabel: "FILE",
        description: "The output file. Determines --lang and --top-level."
    },
    {
        name: "top-level",
        alias: "t",
        type: String,
        typeLabel: "NAME",
        description: "The name for the top level type."
    },
    {
        name: "lang",
        alias: "l",
        type: String,
        typeLabel: langs,
        description: "The target language."
    },
    {
        name: "src-lang",
        alias: "s",
        type: String,
        defaultValue: "json",
        typeLabel: "json|schema",
        description: "The source language (default is json)."
    },
    {
        name: "src",
        type: String,
        multiple: true,
        defaultOption: true,
        typeLabel: "FILE|URL|DIRECTORY",
        description: "The file, url, or data directory to type."
    },
    {
        name: "src-urls",
        type: String,
        typeLabel: "FILE",
        description: "Tracery grammar describing URLs to crawl."
    },
    {
        name: "no-combine-classes",
        type: Boolean,
        description: "Don't combine similar classes."
    },
    {
        name: "no-maps",
        type: Boolean,
        description: "Don't infer maps, always use classes."
    },
    {
        name: "no-enums",
        type: Boolean,
        description: "Don't infer enums, always use strings."
    },
    {
        name: "no-render",
        type: Boolean,
        description: "Don't render output."
    },
    {
        name: "quiet",
        type: Boolean,
        description: "Don't show issues in the generated code."
    },
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Get some help."
    }
];
var sectionsBeforeRenderers = [
    {
        header: "Synopsis",
        content: "$ quicktype [[bold]{--lang} " + langs + "] FILE|URL ..."
    },
    {
        header: "Description",
        content: "Given JSON sample data, quicktype outputs code for working with that data in " + langDisplayNames + "."
    },
    {
        header: "Options",
        optionList: optionDefinitions,
        hide: ["no-render"]
    }
];
var sectionsAfterRenderers = [
    {
        header: "Examples",
        content: [
            chalk.dim("Generate C# to parse a Bitcoin API"),
            "$ quicktype -o LatestBlock.cs https://blockchain.info/latestblock",
            "",
            chalk.dim("Generate Go code from a directory of samples containing:"),
            chalk.dim("  - Foo.json\n  + Bar\n    - bar-sample-1.json\n    - bar-sample-2.json\n  - Baz.url"),
            "$ quicktype -l go samples",
            "",
            chalk.dim("Generate JSON Schema, then TypeScript"),
            "$ quicktype -o schema.json https://blockchain.info/latestblock",
            "$ quicktype -o bitcoin.ts --src-lang schema schema.json"
        ]
    },
    {
        content: "Learn more at [bold]{quicktype.io}"
    }
];
function getTargetLanguage(name) {
    var language = targetLanguages.languageNamed(name);
    if (language) {
        return language;
    }
    console.error("'" + name + "' is not yet supported as an output language.");
    return process.exit(1);
}
function usage() {
    var rendererSections = [];
    _.forEach(targetLanguages.all, function (language) {
        var definitions = language.optionDefinitions;
        if (definitions.length === 0)
            return;
        rendererSections.push({
            header: "Options for " + language.displayName,
            optionList: definitions
        });
    });
    var sections = _.concat(sectionsBeforeRenderers, rendererSections, sectionsAfterRenderers);
    console.log(getUsage(sections));
}
exports.usage = usage;
var graphByInputHash = immutable_1.Map();
var Run = /** @class */ (function () {
    function Run(argv, _doCache) {
        var _this = this;
        this._doCache = _doCache;
        this.getOptionDefinitions = function (opts) {
            return getTargetLanguage(opts.lang).optionDefinitions;
        };
        this.makeGraph = function () {
            var supportsEnums = getTargetLanguage(_this._options.lang).supportsEnums;
            var typeBuilder = new TypeBuilder_1.TypeGraphBuilder();
            if (_this.isInputJSONSchema) {
                immutable_1.Map(_this._allSamples.schemas).forEach(function (schema, name) {
                    typeBuilder.addTopLevel(name, JSONSchemaInput_1.schemaToType(typeBuilder, name, schema));
                });
                return typeBuilder.finish();
            }
            else {
                var doInferMaps = !_this._options.noMaps;
                var doInferEnums = supportsEnums && !_this._options.noEnums;
                var doCombineClasses = !_this._options.noCombineClasses;
                var samplesMap = immutable_1.Map(_this._allSamples.samples);
                var inputs = immutable_1.List([
                    doInferMaps,
                    doInferEnums,
                    doCombineClasses,
                    samplesMap.map(function (values) { return immutable_1.List(values); }),
                    _this._compressedJSON
                ]);
                var inputHash = undefined;
                if (_this._doCache) {
                    inputHash = inputs.hashCode();
                    var maybeGraph = graphByInputHash.get(inputHash);
                    if (maybeGraph !== undefined) {
                        return maybeGraph;
                    }
                }
                var inference_1 = new Inference_1.TypeInference(typeBuilder, doInferMaps, doInferEnums);
                immutable_1.Map(_this._allSamples.samples).forEach(function (cjson, name) {
                    typeBuilder.addTopLevel(name, inference_1.inferType(_this._compressedJSON, name, false, cjson));
                });
                var graph = typeBuilder.finish();
                if (doCombineClasses) {
                    graph = CombineClasses_1.combineClasses(graph);
                }
                if (doInferMaps) {
                    graph = InferMaps_1.inferMaps(graph);
                }
                if (inputHash !== undefined) {
                    graphByInputHash = graphByInputHash.set(inputHash, graph);
                }
                return graph;
            }
        };
        this.renderSamplesOrSchemas = function () {
            var targetLanguage = getTargetLanguage(_this._options.lang);
            try {
                var graph = _this.makeGraph();
                if (_this._options.noRender) {
                    return { lines: ["Done.", ""], annotations: immutable_1.List() };
                }
                return targetLanguage.renderGraphAndSerialize(graph, _this._options.rendererOptions);
            }
            catch (e) {
                console.error(e);
                return process.exit(1);
            }
        };
        this.splitAndWriteJava = function (dir, str) {
            var lines = str.split("\n");
            var filename = null;
            var currentFileContents = "";
            var writeFile = function () {
                if (filename != null) {
                    fs.writeFileSync(path.join(dir, filename), currentFileContents);
                }
                filename = null;
                currentFileContents = "";
            };
            var i = 0;
            while (i < lines.length) {
                var line = lines[i];
                i += 1;
                var results = line.match("^// (.+\\.java)$");
                if (results == null) {
                    currentFileContents += line + "\n";
                }
                else {
                    writeFile();
                    filename = results[1];
                    while (lines[i] === "")
                        i++;
                }
            }
            writeFile();
        };
        this.readSampleFromStream = function (name, readStream) { return __awaiter(_this, void 0, void 0, function () {
            var input, _a, _b, input;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!this.isInputJSONSchema) return [3 /*break*/, 2];
                        _b = (_a = JSON).parse;
                        return [4 /*yield*/, getStream(readStream)];
                    case 1:
                        input = _b.apply(_a, [_c.sent()]);
                        if (_.has(this._allSamples.schemas, name)) {
                            throw new Error("More than one schema given for " + name);
                        }
                        this._allSamples.schemas[name] = input;
                        return [3 /*break*/, 4];
                    case 2: return [4 /*yield*/, this._compressedJSON.readFromStream(readStream)];
                    case 3:
                        input = _c.sent();
                        if (!_.has(this._allSamples.samples, name)) {
                            this._allSamples.samples[name] = [];
                        }
                        this._allSamples.samples[name].push(input);
                        _c.label = 4;
                    case 4: return [2 /*return*/];
                }
            });
        }); };
        this.readSampleFromFileOrUrl = function (name, fileOrUrl) { return __awaiter(_this, void 0, void 0, function () {
            var res;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!fs.existsSync(fileOrUrl)) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.readSampleFromStream(name, fs.createReadStream(fileOrUrl))];
                    case 1:
                        _a.sent();
                        return [3 /*break*/, 5];
                    case 2: return [4 /*yield*/, fetch(fileOrUrl)];
                    case 3:
                        res = _a.sent();
                        return [4 /*yield*/, this.readSampleFromStream(name, res.body)];
                    case 4:
                        _a.sent();
                        _a.label = 5;
                    case 5: return [2 /*return*/];
                }
            });
        }); };
        this.readSampleFromFileOrUrlArray = function (name, filesOrUrls) { return __awaiter(_this, void 0, void 0, function () {
            var _i, filesOrUrls_1, fileOrUrl;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _i = 0, filesOrUrls_1 = filesOrUrls;
                        _a.label = 1;
                    case 1:
                        if (!(_i < filesOrUrls_1.length)) return [3 /*break*/, 4];
                        fileOrUrl = filesOrUrls_1[_i];
                        return [4 /*yield*/, this.readSampleFromFileOrUrl(name, fileOrUrl)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        _i++;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        }); };
        this.readNamedSamplesFromDirectory = function (dataDir) { return __awaiter(_this, void 0, void 0, function () {
            var _this = this;
            var readFilesOrURLsInDirectory, contents, directories, _i, directories_1, dir, sampleName;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        readFilesOrURLsInDirectory = function (d, sampleName) { return __awaiter(_this, void 0, void 0, function () {
                            var files, _i, files_1, file, name, fileOrUrl;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        files = fs
                                            .readdirSync(d)
                                            .map(function (x) { return path.join(d, x); })
                                            .filter(function (x) { return fs.lstatSync(x).isFile(); });
                                        _i = 0, files_1 = files;
                                        _a.label = 1;
                                    case 1:
                                        if (!(_i < files_1.length)) return [3 /*break*/, 4];
                                        file = files_1[_i];
                                        if (sampleName === undefined) {
                                            name = path.basename(file);
                                            sampleName = name.substr(0, name.lastIndexOf("."));
                                        }
                                        fileOrUrl = file;
                                        // If file is a URL string, download it
                                        if (_.endsWith(file, ".url")) {
                                            fileOrUrl = fs.readFileSync(file, "utf8").trim();
                                        }
                                        return [4 /*yield*/, this.readSampleFromFileOrUrl(sampleName, fileOrUrl)];
                                    case 2:
                                        _a.sent();
                                        _a.label = 3;
                                    case 3:
                                        _i++;
                                        return [3 /*break*/, 1];
                                    case 4: return [2 /*return*/];
                                }
                            });
                        }); };
                        contents = fs.readdirSync(dataDir).map(function (x) { return path.join(dataDir, x); });
                        directories = contents.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                        return [4 /*yield*/, readFilesOrURLsInDirectory(dataDir)];
                    case 1:
                        _a.sent();
                        _i = 0, directories_1 = directories;
                        _a.label = 2;
                    case 2:
                        if (!(_i < directories_1.length)) return [3 /*break*/, 5];
                        dir = directories_1[_i];
                        sampleName = path.basename(dir);
                        return [4 /*yield*/, readFilesOrURLsInDirectory(dir, sampleName)];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _i++;
                        return [3 /*break*/, 2];
                    case 5: return [2 /*return*/];
                }
            });
        }); };
        this.runWeb = function (sources) { return __awaiter(_this, void 0, void 0, function () {
            var _i, sources_1, _a, name, samples, _b, samples_1, sample;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        _i = 0, sources_1 = sources;
                        _c.label = 1;
                    case 1:
                        if (!(_i < sources_1.length)) return [3 /*break*/, 6];
                        _a = sources_1[_i], name = _a.name, samples = _a.samples;
                        _b = 0, samples_1 = samples;
                        _c.label = 2;
                    case 2:
                        if (!(_b < samples_1.length)) return [3 /*break*/, 5];
                        sample = samples_1[_b];
                        return [4 /*yield*/, this.readSampleFromStream(name, stringToStream(sample))];
                    case 3:
                        _c.sent();
                        _c.label = 4;
                    case 4:
                        _b++;
                        return [3 /*break*/, 2];
                    case 5:
                        _i++;
                        return [3 /*break*/, 1];
                    case 6: return [2 /*return*/, this.renderSamplesOrSchemas()];
                }
            });
        }); };
        this.run = function () { return __awaiter(_this, void 0, void 0, function () {
            var json, jsonMap, _i, _a, key, exists, directories_2, _b, directories_3, dataDir, filesOrUrls;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        Support_1.assert(!this._options.help, "Cannot print help when run without printing");
                        if (!this._options.srcUrls) return [3 /*break*/, 5];
                        json = JSON.parse(fs.readFileSync(this._options.srcUrls, "utf8"));
                        jsonMap = URLGrammar_1.urlsFromURLGrammar(json);
                        _i = 0, _a = Object.keys(jsonMap);
                        _c.label = 1;
                    case 1:
                        if (!(_i < _a.length)) return [3 /*break*/, 4];
                        key = _a[_i];
                        return [4 /*yield*/, this.readSampleFromFileOrUrlArray(key, jsonMap[key])];
                    case 2:
                        _c.sent();
                        _c.label = 3;
                    case 3:
                        _i++;
                        return [3 /*break*/, 1];
                    case 4: return [3 /*break*/, 13];
                    case 5:
                        if (!(this._options.src.length === 0)) return [3 /*break*/, 7];
                        // FIXME: Why do we have to convert to any here?
                        return [4 /*yield*/, this.readSampleFromStream(this._options.topLevel, process.stdin)];
                    case 6:
                        // FIXME: Why do we have to convert to any here?
                        _c.sent();
                        return [3 /*break*/, 13];
                    case 7:
                        exists = this._options.src.filter(fs.existsSync);
                        directories_2 = exists.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                        _b = 0, directories_3 = directories_2;
                        _c.label = 8;
                    case 8:
                        if (!(_b < directories_3.length)) return [3 /*break*/, 11];
                        dataDir = directories_3[_b];
                        return [4 /*yield*/, this.readNamedSamplesFromDirectory(dataDir)];
                    case 9:
                        _c.sent();
                        _c.label = 10;
                    case 10:
                        _b++;
                        return [3 /*break*/, 8];
                    case 11:
                        filesOrUrls = this._options.src.filter(function (x) { return !_.includes(directories_2, x); });
                        if (!!_.isEmpty(filesOrUrls)) return [3 /*break*/, 13];
                        return [4 /*yield*/, this.readSampleFromFileOrUrlArray(this._options.topLevel, filesOrUrls)];
                    case 12:
                        _c.sent();
                        _c.label = 13;
                    case 13: return [2 /*return*/, this.renderSamplesOrSchemas()];
                }
            });
        }); };
        this.runAndPrint = function () { return __awaiter(_this, void 0, void 0, function () {
            var _a, lines, annotations, output;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if (this._options.help) {
                            usage();
                            return [2 /*return*/];
                        }
                        return [4 /*yield*/, this.run()];
                    case 1:
                        _a = _b.sent(), lines = _a.lines, annotations = _a.annotations;
                        output = lines.join("\n");
                        if (this._options.out) {
                            if (this._options.lang === "java") {
                                this.splitAndWriteJava(path.dirname(this._options.out), output);
                            }
                            else {
                                fs.writeFileSync(this._options.out, output);
                            }
                        }
                        else {
                            process.stdout.write(output);
                        }
                        if (this._options.quiet) {
                            return [2 /*return*/];
                        }
                        annotations.forEach(function (sa) {
                            var annotation = sa.annotation;
                            if (!(annotation instanceof Annotation_1.IssueAnnotationData))
                                return;
                            var lineNumber = sa.span.start.line;
                            var humanLineNumber = lineNumber + 1;
                            console.error("\nIssue in line " + humanLineNumber + ": " + annotation.message);
                            console.error(humanLineNumber + ": " + lines[lineNumber]);
                        });
                        return [2 /*return*/];
                }
            });
        }); };
        // Parse the options in argv and split them into global options and renderer options,
        // according to each option definition's `renderer` field.  If `partial` is false this
        // will throw if it encounters an unknown option.
        this.parseOptions = function (definitions, argv, partial) {
            var opts = commandLineArgs(definitions, {
                argv: argv,
                partial: partial
            });
            var options = { rendererOptions: {} };
            definitions.forEach(function (o) {
                if (!(o.name in opts))
                    return;
                var v = opts[o.name];
                if (o.renderer)
                    options.rendererOptions[o.name] = v;
                else {
                    var k = _.lowerFirst(o.name
                        .split("-")
                        .map(_.upperFirst)
                        .join(""));
                    options[k] = v;
                }
            });
            return _this.inferOptions(options);
        };
        this.inferOptions = function (opts) {
            return __assign({ src: opts.src || [], srcLang: opts.srcLang || "json", lang: opts.lang || _this.inferLang(opts), topLevel: opts.topLevel || _this.inferTopLevel(opts), noMaps: !!opts.noMaps, noEnums: !!opts.noEnums, noCombineClasses: !!opts.noCombineClasses, noRender: !!opts.noRender, help: !!opts.help, quiet: !!opts.quiet }, opts);
        };
        this.inferLang = function (options) {
            // Output file extension determines the language if language is undefined
            if (options.out) {
                var extension = path.extname(options.out);
                if (extension === "") {
                    console.error("Please specify a language (--lang) or an output file extension.");
                    process.exit(1);
                }
                return extension.substr(1);
            }
            return "go";
        };
        this.inferTopLevel = function (options) {
            // Output file name determines the top-level if undefined
            if (options.out) {
                var extension = path.extname(options.out);
                var without = path.basename(options.out).replace(extension, "");
                return without;
            }
            // Source determines the top-level if undefined
            if (options.src && options.src.length === 1) {
                var src = options.src[0];
                var extension = path.extname(src);
                var without = path.basename(src).replace(extension, "");
                return without;
            }
            return "TopLevel";
        };
        if (_.isArray(argv)) {
            // We can only fully parse the options once we know which renderer is selected,
            // because there are renderer-specific options.  But we only know which renderer
            // is selected after we've parsed the options.  Hence, we parse the options
            // twice.  This is the first parse to get the renderer:
            var incompleteOptions = this.parseOptions(optionDefinitions, argv, true);
            var rendererOptionDefinitions = this.getOptionDefinitions(incompleteOptions);
            // Use the global options as well as the renderer options from now on:
            var allOptionDefinitions = _.concat(optionDefinitions, rendererOptionDefinitions);
            try {
                // This is the parse that counts:
                this._options = this.parseOptions(allOptionDefinitions, argv, false);
            }
            catch (error) {
                if (error.name === "UNKNOWN_OPTION") {
                    console.error("Error: Unknown option");
                    usage();
                    process.exit(1);
                }
                throw error;
            }
        }
        else {
            this._options = this.inferOptions(argv);
        }
        this._compressedJSON = new CompressedJSON_1.CompressedJSON();
        this._allSamples = { samples: {}, schemas: {} };
    }
    Object.defineProperty(Run.prototype, "isInputJSONSchema", {
        get: function () {
            return this._options.srcLang === "schema";
        },
        enumerable: true,
        configurable: true
    });
    return Run;
}());
exports.Run = Run;
